/* $NetBSD: $ */

/*-
 * Copyright (c) 2017 Jared McNeill <jmcneill@invisible.ca>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "opt_soc.h"
#include "opt_multiprocessor.h"
#include "opt_console.h"

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: $");

#include <sys/param.h>
#include <sys/bus.h>
#include <sys/cpu.h>
#include <sys/device.h>
#include <sys/termios.h>

#include <dev/fdt/fdtvar.h>
#include <arm/fdt/arm_fdtvar.h>

#include <uvm/uvm_extern.h>

#include <machine/bootconfig.h>
#include <arm/cpufunc.h>

#include <arm/cortex/a9tmr_var.h>
#include <arm/cortex/pl310_var.h>
#include <arm/cortex/gic_reg.h>

#include <arm/zynq/zynq_platform.h>
#include <arm/zynq/zynq_uartreg.h>

#include <libfdt.h>

#define	ZYNQ_REF_FREQ	24000000

#define	SUN4I_TIMER_BASE	0x01c20c00
#define	SUN4I_TIMER_SIZE	0x90
#define	SUN4I_TIMER_1_CTRL	0x20
#define	 SUN4I_TIMER_1_CTRL_CLK_SRC	__BITS(3,2)
#define	 SUN4I_TIMER_1_CTRL_CLK_SRC_OSC24M	1
#define	 SUN4I_TIMER_1_CTRL_RELOAD	__BIT(1)
#define	 SUN4I_TIMER_1_CTRL_EN		__BIT(0)
#define	SUN4I_TIMER_1_INTV_VALUE 0x24
#define	SUN4I_TIMER_1_VAL	0x28

#define	SUN4I_WDT_BASE		0x01c20c90
#define	SUN4I_WDT_SIZE		0x10
#define	SUN4I_WDT_CTRL		0x00
#define	 SUN4I_WDT_CTRL_KEY	(0x333 << 1)
#define	 SUN4I_WDT_CTRL_RESTART	__BIT(0)
#define	SUN4I_WDT_MODE		0x04
#define	 SUN4I_WDT_MODE_RST_EN	__BIT(1)
#define	 SUN4I_WDT_MODE_EN	__BIT(0)

extern struct bus_space arm_generic_bs_tag;
extern struct bus_space arm_generic_a4x_bs_tag;
extern struct arm32_bus_dma_tag arm_generic_dma_tag;

#define zynq_dma_tag		arm_generic_dma_tag
#define zynq_bs_tag		arm_generic_bs_tag
#define zynq_a4x_bs_tag		arm_generic_a4x_bs_tag

void zynq_platform_early_putchar(char);

static const struct pmap_devmap *
zynq_platform_devmap(void)
{
	static const struct pmap_devmap devmap[] = {
		DEVMAP_ENTRY(ZYNQ_IOREG_VBASE,
			     ZYNQ_IOREG_PBASE,
			     ZYNQ_IOREG_SIZE),
		DEVMAP_ENTRY(ZYNQ_GPV_VBASE,
			     ZYNQ_GPV_PBASE,
			     ZYNQ_GPV_SIZE),
		DEVMAP_ENTRY(ZYNQ_ARMCORE_VBASE,
			     ZYNQ_ARMCORE_PBASE,
			     ZYNQ_ARMCORE_SIZE),
		DEVMAP_ENTRY_END
	};

	return devmap;
}

static void
zynq_platform_init_attach_args(struct fdt_attach_args *faa)
{
	faa->faa_bst = &zynq_bs_tag;
	faa->faa_a4x_bst = &zynq_a4x_bs_tag;
	faa->faa_dmat = &zynq_dma_tag;
}

void
zynq_platform_early_putchar(char c)
{
#ifdef CONSADDR
#define CONSADDR_VA     ((CONSADDR - ZYNQ_IOREG_PBASE) + ZYNQ_IOREG_VBASE)
	volatile uint32_t *uartaddr = cpu_earlydevice_va_p() ?
	    (volatile uint32_t *)CONSADDR_VA :
	    (volatile uint32_t *)CONSADDR;

	while ((le32toh(uartaddr[UART_CHANNEL_STS/4]) & STS_TEMPTY) == 0)
		;

	uartaddr[UART_TX_RX_FIFO/4] = htole32(c);
#endif
}

static void
zynq_platform_device_register(device_t dev, void *aux)
{
	prop_dictionary_t dict = device_properties(dev);

	if (device_is_a(dev, "arma9tmr")) {
		prop_dictionary_set_uint32(dict, "frequency",
			ZYNQ_REF_FREQ / 4);
	}
}

static u_int
zynq_platform_uart_freq(void)
{
	return ZYNQ_REF_FREQ;
}

static void
zynq_platform_bootstrap(void)
{
	void *fdt_data = __UNCONST(fdtbus_get_data());
	const int chosen_off = fdt_path_offset(fdt_data, "/chosen");
	if (chosen_off < 0)
		return;

	if (match_bootconf_option(boot_args, "console", "fb")) {
		const int framebuffer_off =
		    fdt_path_offset(fdt_data, "/chosen/framebuffer");
		if (framebuffer_off >= 0) {
			const char *status = fdt_getprop(fdt_data,
			    framebuffer_off, "status", NULL);
			if (status == NULL || strncmp(status, "ok", 2) == 0) {
				fdt_setprop_string(fdt_data, chosen_off,
				    "stdout-path", "/chosen/framebuffer");
			}
		}
	} else if (match_bootconf_option(boot_args, "console", "serial")) {
		fdt_setprop_string(fdt_data, chosen_off,
		    "stdout-path", "serial0:115200n8");
	}

	arm_fdt_cpu_bootstrap();
}

static void
zynq_platform_reset(void)
{
#if 0
	bus_space_tag_t bst = &zynq_bs_tag;
	bus_space_handle_t bsh;

	bus_space_map(bst, SUN4I_WDT_BASE, SUN4I_WDT_SIZE, 0, &bsh);

	bus_space_write_4(bst, bsh, SUN4I_WDT_CTRL,
	    SUN4I_WDT_CTRL_KEY | SUN4I_WDT_CTRL_RESTART);
	for (;;) {
		bus_space_write_4(bst, bsh, SUN4I_WDT_MODE,
		    SUN4I_WDT_MODE_EN | SUN4I_WDT_MODE_RST_EN);
	}
#endif
}

static const struct arm_platform zynq_platform = {
	.ap_devmap = zynq_platform_devmap,
	.ap_bootstrap = zynq_platform_bootstrap,
	.ap_init_attach_args = zynq_platform_init_attach_args,
	.ap_device_register = zynq_platform_device_register,
	.ap_reset = zynq_platform_reset,
	.ap_delay = a9tmr_delay,
	.ap_uart_freq = zynq_platform_uart_freq,
#if 0
	.ap_mpstart = arm_fdt_cpu_mpstart,
#endif
};

ARM_PLATFORM(parallella, "adapteva,parallella", &zynq_platform);
ARM_PLATFORM(microzed, "xlnx,zynq-microzed", &zynq_platform);
ARM_PLATFORM(zc702, "xlnx,zynq-zc702", &zynq_platform);
ARM_PLATFORM(zc706, "xlnx,zynq-zc706", &zynq_platform);
ARM_PLATFORM(zed, "xlnx,zynq-zed", &zynq_platform);
ARM_PLATFORM(zybo, "digilent,zynq-zybo", &zynq_platform);
ARM_PLATFORM(zybo_z7, "digilent,zynq-zybo-z7", &zynq_platform);

